#include <stdarg.h>
#include <stdint-gcc.h>
#include "double_to_str.h"

extern int mvsnprintf (char *pStr, int length, const char * pFormat, va_list ap);

int msnprintf (char * buf, int length, const char *fmt, ...) {
  va_list    ap;
  signed int rc;

  va_start (ap, fmt);
  rc = mvsnprintf (buf, length, fmt, ap);
  va_end (ap);

  return rc;
}

//------------------------------------------------------------------------------
// Writes a character inside the given string. Returns 1.
// \param pStr  Storage string.
// \param c  Character to write.
//------------------------------------------------------------------------------
static signed int PutChar (char *pStr, char c) {
  *pStr = c;
  return 1;
}
//------------------------------------------------------------------------------
// Writes a string inside the given string.
// Returns the size of the written
// string.
// \param pStr  Storage string.
// \param pSource  Source string.
//------------------------------------------------------------------------------
static signed int PutString (char *pStr, const char *pSource) {
  signed int num = 0;

  while (*pSource != 0) {

    *pStr++ = *pSource++;
    num++;
  }

  return num;
}

//------------------------------------------------------------------------------
// Writes an unsigned int inside the given string, using the provided fill &
// width parameters.
// Returns the size in characters of the written integer.
// \param pStr  Storage string.
// \param fill  Fill character.
// \param width  Minimum integer width.
// \param value  Integer value.
//------------------------------------------------------------------------------
static signed int PutUnsignedInt (
  char *pStr,
  char fill,
  signed int width,
  unsigned int value) {
  signed int num = 0;

  // Take current digit into account when calculating width
  width--;

  // Recursively write upper digits
  if ( (value / 10) > 0) {

    num = PutUnsignedInt (pStr, fill, width, value / 10);
    pStr += num;
  }
  // Write filler characters
  else {

    while (width > 0) {

      PutChar (pStr, fill);
      pStr++;
      num++;
      width--;
    }
  }

  // Write lower digit
  num += PutChar (pStr, (value % 10) + '0');

  return num;
}
//------------------------------------------------------------------------------
static signed int PutUnsignedOct (
  char *pStr,
  char fill,
  signed int width,
  unsigned int value) {
  signed int num = 0;
  // Take current digit into account when calculating width
  width--;
  // Recursively write upper digits
  if ( (value / 8) > 0) {
    num = PutUnsignedOct (pStr, fill, width, value / 8);
    pStr += num;
  }
  // Write filler characters
  else {
    while (width > 0) {
      PutChar (pStr, fill);
      pStr++;
      num++;
      width--;
    }
  }
  // Write lower digit
  num += PutChar (pStr, (value % 8) + '0');
  return num;
}

//------------------------------------------------------------------------------
// Writes a signed int inside the given string, using the provided fill & width
// parameters.
// Returns the size of the written integer.
// \param pStr  Storage string.
// \param fill  Fill character.
// \param width  Minimum integer width.
// \param value  Signed integer value.
//------------------------------------------------------------------------------
static signed int PutSignedInt (
  char *pStr,
  char fill,
  signed int width,
  signed int value) {
  signed int num = 0;
  unsigned int absolute;

  // Compute absolute value
  if (value < 0) {

    absolute = -value;
  } else {

    absolute = value;
  }

  // Take current digit into account when calculating width
  width--;

  // Recursively write upper digits
  if ( (absolute / 10) > 0) {

    if (value < 0) {

      num = PutSignedInt (pStr, fill, width, - (absolute / 10));
    } else {

      num = PutSignedInt (pStr, fill, width, absolute / 10);
    }
    pStr += num;
  } else {

    // Reserve space for sign
    if (value < 0) {

      width--;
    }

    // Write filler characters
    while (width > 0) {

      PutChar (pStr, fill);
      pStr++;
      num++;
      width--;
    }

    // Write sign
    if (value < 0) {

      num += PutChar (pStr, '-');
      pStr++;
    }
  }

  // Write lower digit
  num += PutChar (pStr, (absolute % 10) + '0');

  return num;
}

//------------------------------------------------------------------------------
// Writes an hexadecimal value into a string, using the given fill, width &
// capital parameters.
// Returns the number of char written.
// \param pStr  Storage string.
// \param fill  Fill character.
// \param width  Minimum integer width.
// \param maj  Indicates if the letters must be printed in lower- or upper-case.
// \param value  Hexadecimal value.
//------------------------------------------------------------------------------
static signed int PutHexa (
  char *pStr,
  char fill,
  signed int width,
  unsigned char maj,
  unsigned int value) {
  signed int num = 0;

  // Decrement width
  width--;

  // Recursively output upper digits
  if ( (value >> 4) > 0) {

    num += PutHexa (pStr, fill, width, maj, value >> 4);
    pStr += num;
  }
  // Write filler chars
  else {

    while (width > 0) {

      PutChar (pStr, fill);
      pStr++;
      num++;
      width--;
    }
  }

  // Write current digit
  if ( (value & 0xF) < 10) {

    PutChar (pStr, (value & 0xF) + '0');
  } else if (maj) {

    PutChar (pStr, (value & 0xF) - 10 + 'A');
  } else {

    PutChar (pStr, (value & 0xF) - 10 + 'a');
  }
  num++;

  return num;
}
/*
static signed int PutDouble (char * pStr, const double value) {
  signed int num = 0;
  union ud_u {
    double   x;
    uint64_t u;
  } ud;
  ud.x = value;
  uint64_t tu = ud.u;
  const char * hexstr = "0123456789ABCDEF";
  char * ptr = pStr + 16;
  for (unsigned n=0; n<16; n++) {
    unsigned i = tu & 0xFu;
    tu >>= 4;
    *--ptr = hexstr [i];
    num += 1;
  }
  pStr [num] = '\0';
  return num;
}
*/
/// Stores the result of a formatted string into another string. Format
/// arguments are given in a va_list instance.
/// Return the number of characters written.
/// \param pStr    Destination string.
/// \param length  Length of Destination string.
/// \param pFormat Format string.
/// \param ap      Argument list.
//------------------------------------------------------------------------------
int mvsnprintf (char *pStr, int length, const char * pFormat, va_list ap) {
  char          fill;
  unsigned char width;
  signed int    num = 0;
  signed int    size = 0;
  // Clear the string
  if (pStr) {
    *pStr = 0;
  }
  // Phase string
  while (*pFormat != 0 && size < length) {
    // Normal character
    if (*pFormat != '%') {
      *pStr++ = *pFormat++;
      size++;
    }
    // Escaped '%'
    else if (* (pFormat+1) == '%') {
      *pStr++ = '%';
      pFormat += 2;
      size++;
    }
    // Token delimiter
    else {
      fill = ' ';
      width = 0;
      pFormat++;
      // Parse filler
      if (*pFormat == '0') {
        fill = '0';
        pFormat++;
      }
      // Parse width
      while ( (*pFormat >= '0') && (*pFormat <= '9')) {
        unsigned char dwidth = (width << 3) + (width << 1); // 10 * width = 8 * width + 2 * width
        width = (dwidth) + *pFormat-'0';
        pFormat++;
      }
      // Check if there is enough space
      if (size + width > length) {
        width = length - size;
      }

      // Parse type
      switch (*pFormat) {
      case 'd':
      case 'i': num = PutSignedInt   (pStr, fill, width, va_arg (ap, signed int));   break;
      case 'u': num = PutUnsignedInt (pStr, fill, width, va_arg (ap, unsigned int)); break;
      case 'x': num = PutHexa (pStr, fill, width, 0, va_arg (ap, unsigned int));     break;
      case 'X': num = PutHexa (pStr, fill, width, 1, va_arg (ap, unsigned int));     break;
      case 's': num = PutString (pStr, va_arg (ap, char *));       break;
      case 'c': num = PutChar   (pStr, va_arg (ap, unsigned int)); break;
      case 'o': num = PutUnsignedOct (pStr, fill, width, va_arg (ap, unsigned int)); break;
#if 1
      case 'E':
      case 'F':
      case 'G':
      case 'e':
      case 'f':
//    case 'g': num = PutDouble (pStr, va_arg (ap, double)); break;
      case 'g': num = double_to_str (pStr, va_arg (ap, double)); break;
#endif
      default:
        return 0;
      }
      pFormat++;
      pStr += num;
      size += num;
    }
  }

  // NULL-terminated (final \0 is not counted)
  if (size < length) {
    *pStr = 0;
  } else {
    * (--pStr) = 0;
    size--;
  }

  return size;
}

